@@ -28,6 +28,45 @@
 #include <Python.h>
 #include <fuse.h>
 #include <sys/ioctl.h>
+#ifndef _UAPI_ASM_GENERIC_IOCTL_H
+/* Essential IOCTL definitions from Linux /include/uapi/asm-generic/ioctl.h
+   to fix compilation errors on FreeBSD 
+   Mikhail Zakharov <zmey20000@thoo.com> 2018.10.22 */
+
+#define _IOC_NRBITS     8
+#define _IOC_TYPEBITS   8
+
+#ifndef _IOC_SIZEBITS
+# define _IOC_SIZEBITS  14
+#endif
+
+#ifndef _IOC_DIRBITS
+# define _IOC_DIRBITS   2
+#endif
+
+#define _IOC_SIZEMASK   ((1 << _IOC_SIZEBITS)-1)
+#define _IOC_DIRMASK    ((1 << _IOC_DIRBITS)-1)
+
+#define _IOC_NRSHIFT    0
+#define _IOC_TYPESHIFT  (_IOC_NRSHIFT+_IOC_NRBITS)
+#define _IOC_SIZESHIFT  (_IOC_TYPESHIFT+_IOC_TYPEBITS)
+#define _IOC_DIRSHIFT   (_IOC_SIZESHIFT+_IOC_SIZEBITS)
+
+#ifndef _IOC_NONE
+# define _IOC_NONE      0U
+#endif
+
+#ifndef _IOC_WRITE
+# define _IOC_WRITE     1U
+#endif
+
+#ifndef _IOC_READ
+# define _IOC_READ      2U
+#endif
+
+#define _IOC_DIR(nr)            (((nr) >> _IOC_DIRSHIFT) & _IOC_DIRMASK)
+#define _IOC_SIZE(nr)           (((nr) >> _IOC_SIZESHIFT) & _IOC_SIZEMASK)
+#endif
 
 
 #ifndef FUSE_VERSION
@@ -58,7 +97,8 @@ static PyObject *getattr_cb=NULL, *readlink_cb=NULL, *readdir_cb=NULL,
   *releasedir_cb=NULL, *fsyncdir_cb=NULL, *flush_cb=NULL, *ftruncate_cb=NULL,
   *fgetattr_cb=NULL, *getxattr_cb=NULL, *listxattr_cb=NULL, *setxattr_cb=NULL,
   *removexattr_cb=NULL, *access_cb=NULL, *lock_cb = NULL, *utimens_cb = NULL,
-  *bmap_cb = NULL, *fsinit_cb=NULL, *fsdestroy_cb = NULL, *ioctl_cb = NULL;
+  *bmap_cb = NULL, *fsinit_cb=NULL, *fsdestroy_cb = NULL, *ioctl_cb = NULL,
+  *poll_cb = NULL;
 
 
 static PyObject *Py_FuseError;
@@ -517,12 +557,22 @@ read_func(const char *path, char *buf, size_t s, off_t off)
 	PROLOGUE( PYO_CALLWITHFI(fi, read_cb, snK, path, s, off) )
 #endif
 
+
+#if PY_MAJOR_VERSION >= 3
+	if(PyBytes_Check(v)) {
+		if(PyBytes_Size(v) > s)
+			goto OUT_DECREF;
+		memcpy(buf, PyBytes_AsString(v), PyBytes_Size(v));
+		ret = PyBytes_Size(v);
+	}
+#else
 	if(PyString_Check(v)) {
 		if(PyString_Size(v) > s)
 			goto OUT_DECREF;
 		memcpy(buf, PyString_AsString(v), PyString_Size(v));
 		ret = PyString_Size(v);
 	}
+#endif
 
 	EPILOGUE
 }
@@ -536,7 +586,11 @@ static int
 write_func(const char *path, const char *buf, size_t t, off_t off)
 #endif
 {
+#if PY_MAJOR_VERSION >= 3
+	PROLOGUE( PYO_CALLWITHFI(fi, write_cb, sy#K, path, buf, t, off) )
+#else
 	PROLOGUE( PYO_CALLWITHFI(fi, write_cb, ss#K, path, buf, t, off) )
+#endif
 	EPILOGUE
 }
 
@@ -991,6 +1045,44 @@ ioctl_func(const char *path, int cmd, void *arg,
 	}
 	EPILOGUE
 }
+
+static const char pollhandle_name[] = "pollhandle";
+
+static void
+pollhandle_destructor(PyObject *p)
+{
+	struct fuse_pollhandle *ph;
+
+	ph = PyCapsule_GetPointer(p, pollhandle_name);
+	fuse_pollhandle_destroy(ph);
+}
+
+static int
+poll_func(const char *path, struct fuse_file_info *fi,
+	  struct fuse_pollhandle *ph, unsigned *reventsp)
+{
+	PyObject *pollhandle = Py_None;
+
+	if (ph)
+		pollhandle = PyCapsule_New(ph, pollhandle_name, pollhandle_destructor);
+
+	PROLOGUE(PYO_CALLWITHFI(fi, poll_cb, sO, path, pollhandle));
+
+OUT_DECREF:
+	Py_DECREF(v);
+OUT:
+	if (ph)
+		Py_DECREF(pollhandle);
+
+	PYUNLOCK();
+
+	if (ret > 0) {
+		*reventsp = ret;
+		ret = 0;
+	}
+
+	return ret;
+}
 #endif
 
 static int
@@ -1036,13 +1128,14 @@ Fuse_main(PyObject *self, PyObject *args, PyObject *kw)
 		"create", "opendir", "releasedir", "fsyncdir", "flush",
 	        "ftruncate", "fgetattr", "getxattr", "listxattr", "setxattr",
 	        "removexattr", "access", "lock", "utimens", "bmap",
-		"fsinit", "fsdestroy", "ioctl", "fuse_args", "multithreaded", NULL
+		"fsinit", "fsdestroy", "ioctl",  "poll", "fuse_args",
+		"multithreaded", NULL
 	};
 
 	memset(&op, 0, sizeof(op));
 
 	if (!PyArg_ParseTupleAndKeywords(args, kw,
-	                                 "|OOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOi",
+	                                 "|OOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOi",
 	                                 kwlist, &getattr_cb, &readlink_cb,
 	                                 &readdir_cb, &mknod_cb, &mkdir_cb,
 	                                 &unlink_cb, &rmdir_cb, &symlink_cb,
@@ -1058,7 +1151,7 @@ Fuse_main(PyObject *self, PyObject *args, PyObject *kw)
 	                                 &removexattr_cb, &access_cb,
 	                                 &lock_cb, &utimens_cb, &bmap_cb,
 	                                 &fsinit_cb, &fsdestroy_cb, &ioctl_cb,
-	                                 &fargseq, &multithreaded))
+	                                 &poll_cb, &fargseq, &multithreaded))
 		return NULL;
 
 #define DO_ONE_ATTR_AS(fname, pyname)		\
@@ -1120,6 +1213,7 @@ Fuse_main(PyObject *self, PyObject *args, PyObject *kw)
 #endif
 #if FUSE_VERSION >= 28
 	DO_ONE_ATTR(ioctl);
+	DO_ONE_ATTR(poll);
 #endif
 
 #undef DO_ONE_ATTR
@@ -1282,11 +1376,32 @@ FuseAPIVersion(PyObject *self, PyObject *args)
 	return favers;
 }
 
+static const char FuseNotifyPoll__doc__[] = "Notify IO readiness event.";
+
+static PyObject *
+FuseNotifyPoll(PyObject *self, PyObject *arg)
+{
+	struct fuse_pollhandle *ph;
+	int ret;
+
+	ph = PyCapsule_GetPointer(arg, pollhandle_name);
+	if (!ph) {
+		PyErr_SetString(PyExc_TypeError,
+		                "pollhandle is not a FUSE poll handle");
+		return NULL;
+	}
+
+	ret = fuse_notify_poll(ph);
+
+	return PyInt_FromLong(ret);
+}
+
 static PyMethodDef Fuse_methods[] = {
 	{"main",	(PyCFunction)Fuse_main,	 METH_VARARGS|METH_KEYWORDS},
 	{"FuseGetContext", (PyCFunction)FuseGetContext, METH_VARARGS, FuseGetContext__doc__},
 	{"FuseInvalidate", (PyCFunction)FuseInvalidate, METH_VARARGS, FuseInvalidate__doc__},
 	{"FuseAPIVersion", (PyCFunction)FuseAPIVersion, METH_NOARGS,  FuseAPIVersion__doc__},
+	{"FuseNotifyPoll", (PyCFunction)FuseNotifyPoll, METH_O,       FuseNotifyPoll__doc__},
 	{NULL,		NULL}		/* sentinel */
 };
 
